// File: ccs_compact_float.h -*- c++ -*-
// Time-stamp: <14 Februar 2008 12:33:07  CET --  suvrit>
// Impl. of ccs format sparse matrices

#ifndef CCS_COMPACT_FLOAT_H
#define CCS_COMPACT_FLOAT_H

#include "sparsematrix_compact_float.h"

namespace SSLib {
  class crs_compact_float;
  /**
  * @class ccs_compact_float, contains an implementation of a CCS sparsematrix. It supports
  * many of the operations suggested by the abstract class sparsematrix_compact_float
  * Basically, this class tries to minimize the storage used ---
  * the colptrs have to be size_t, because they index into the m_rowindx
  * array, which can be upto 2^32 elements long. The m_rowindx array is long,
  * but each element is a row index, between 0 and 65535 -- so we spend only 2
  * bytes on it. Naturally, the values are stored here as floats...so 4 bytes
  * per value. For highly specialized matrices, where for e.g., one knows the
  * min and max values, one can develop a tighter storage, but then at the
  * cost of more complicated operations ....
  *
  */
  class ccs_compact_float : public sparsematrix_compact_float {
  protected:			// for efficiency reasons.
    /**
    * The actual non-zeroes themselves.
    */
    float* m_val;

    /**
    * Colptrs for CCS structure. These hold indices into the
    * <code>m_rowindx</code> and <code>val</code>. The i-th column starts at
    * m_colptrs[i] and ends at m_colptrs[i+1]-1
    */
    size_t* m_colptrs;
  

    /**
    * Row indices for CCs structure. This array contain the row indices of
    * the matrix elements. For example, m_rowindx[ m_colptrs[i] ] gives the
    * row index of the 1st non-zero element in the i-th column.
    */

    typedef unsigned short int index_t;
    index_t* m_rowindx;

  protected:    
    /**
    * Is the data in externally allocated arrays. That is, if the m_val,
    * m_colptrs and m_rowindx arrays are externally allocated (useful for
    * creating a ccs sparsematrix from another matrix of similar structure etc.
    */
    bool m_isExternal;

    /**
    * Filename prefix. A CCS matrix is loaded from a collection of 4
    * files. For example, if m_filename="mat", then the code that loads this
    * matrix will search for the files 'mat_dim', 'mat_row_ccs',
    * 'mat_col_ccs' and 'mat_<m_type>_nz' depending on the value of <code>m_type</code>
    */
    std::string m_filename;

    /* matrix type, 'txx', 'tfn', etc. */
    std::string m_type;

  private:
    /* load files */
    virtual int load_bin(std::string prefix);
    virtual int load_txt  (std::string prefix);
    virtual int save_bin(const char*); 
    virtual int save_txt  (const char*);

  public:
    ccs_compact_float () {
      m_isExternal = false;
      m_filename = "";
      m_type = "txx";
      m_val = 0;
      m_colptrs = 0;
      m_rowindx = 0;
    }

    virtual ~ccs_compact_float () {
      //free_data();
    }

    
    void free_data();
    

    std::string get_type() const { return m_type; }
    void set_type (std::string s) { m_type = s;}

    /** @name MatrixCreation
    * Creating and saving ccs matrix functions. These functions allow the
    * user to load a CCS matrix from file, or from already available
    * arrays. They also allow the user to save the current matrix to file.
    *
    */

    /*@{*/


    // Allocate the matrix
    int matrix_alloc (size_t m, size_t n, size_t nz);
    int matrix_calloc (size_t m, size_t n, size_t nz);
    int matrix_resize (size_t m, size_t n, size_t nz);

    virtual int load(const char*, bool asbin);
    virtual int save_coord_bin(const char*);
    virtual int save (const char*, bool asbin);

    /// Write out matrix in a specified way
    virtual int    save(const char*, bool asbin, int typ);
    
    /// Write out matrix to disk in CCS, CRS, etc. formats
    virtual int save_as_crs (const char*, bool);
    virtual int save_as_coord (const char*, bool);
    virtual int save_as_colcoord (const char*, bool);

    virtual int save_as_rowcoord (const char*, bool);
    virtual int save_rowcoord_bin(const char*);
    virtual int save_rowcoord_txt(const char*);

    virtual int save_as_matlab (const char*, bool);
    virtual int save_matlab_bin(const char* fil);
    virtual int save_matlab_txt(const char*);

    virtual int save_as_hb (const char*, bool);
    virtual int save_hb_bin(const char*);
    virtual int save_hb_txt(const char*);

    virtual int save_as_dense (const char*, bool);
    virtual int save_dense_bin(const char*);
    virtual int save_dense_txt(const char*);

    void externally_alloc (size_t* col, index_t* row, float* val);
    // Return a deep copy
    ccs_compact_float* clone();

    // Augment matrix by adding a row
    virtual int matrix_add_row(vector_float* r);
    virtual int matrix_remove_row(size_t r);

    // Augment by adding a column
    virtual int matrix_add_col(vector_float* c);
    virtual int matrix_remove_col(size_t c);

    /*@}*/


    int     scale(float);
    int     add_const(float);

    /** @name Accessors
    * Accessing the matrix elements 
    */
    /*@{*/

    size_t* get_colptrs() { return m_colptrs;}
    index_t* get_rowindx() { return m_rowindx;}
    float* get_data() { return m_val;}

    /// Return row 'r' of matrix
    int get_row(size_t r, vector_float*&);
    /// Return col 'c' of matrix
    int get_col(size_t c, vector_float*&);
    /// return the (i, j) entry of the matrix
    float  get (size_t i, size_t j);
    float  get (size_t i, size_t j, size_t* posn);
    float  operator()(size_t i, size_t j);
    /// set the (i, j) entry of the matrix.
    int     set (size_t i, size_t j, float val);
    int     set (size_t posn, float val);
    /// Sets the specified row to the given vector
    virtual int set_row(size_t r, vector_float*&);

    /// Sets the existing non-zeros, ignores other entries
    virtual int fast_set_row(size_t r, vector_float*&);


    /// Sets the specified col to the given vector
    virtual int set_col(size_t c, vector_float*&);

    /// Sets only the existing non-zeros, ignores others
    virtual int fast_set_col(size_t c, vector_float*&);

    /// Returns main or second diagonal (if p == true)
    virtual int get_diag(bool, vector_float*&);

    /// NOP
    virtual int set_diag(bool b, vector_float*& v) {
      return -1;
    }

    /// Returns a submatrix that goes from row i1--i2 and cols j1--j2
    virtual int submatrix(size_t i1, size_t j1, size_t i2, size_t j2, matrix_compact_float*&);
    
    /// Returns a submatrix given the row-col index sets
    virtual int submatrix(gsl_vector_uint* I, gsl_vector_uint* J, matrix_compact_float*& m) {
      return -1;
    }

    /*@}*/

    /** @name Elementwise
    * Elementwise functions/operations on this matrix.
    */
    /*@{*/

    // Vector l_p norms for this matrix
    virtual float norm (float p);
    virtual float norm (char*  p);
    // Apply an arbitrary function elementwise to this matrix
    virtual float apply (float (* fn)(float));

    /*@}*/

    /*@{*/
    virtual float col_norm (size_t c);
    virtual float col_norm (size_t c, float p);
    virtual float col_dot  (size_t c1, size_t c2) ;
    virtual float col_dot  (size_t c, vector_float* v);
    virtual int    col_add  (size_t c1, size_t c2) ;
    virtual int    col_sub  (size_t c1, size_t c2) ;
    virtual int    col_scale(size_t c1, float s);
    virtual int    col_sub  (size_t c1, vector_float* v);
    virtual int    col_add  (size_t c1, vector_float* v) ;
    virtual int    col_scale(size_t c1, float s, vector_float* r);
    virtual int    col_sum  (vector_float*);
    /*@}*/

    
    /** @name
    * Operations on rows of this matrix
    */
    /*@{*/
    virtual float row_norm (size_t r);
    virtual float row_norm (size_t r, float p);
    virtual float row_dot  (size_t r1, size_t r2) ;
    virtual float row_dot  (size_t r, vector_float* v);
    virtual float row_dot  (size_t r, vector_float* x, size_t* ix,size_t);
    virtual int    row_add  (size_t r1, size_t r2) ;
    virtual int    row_sub  (size_t r1, size_t r2) ;
    virtual int    row_scale(size_t c1, float s);
    virtual int    row_add  (size_t r1, vector_float* r);
    virtual int    row_sub  (size_t r1, vector_float* r);
    virtual int row_scale(size_t c1, float s, vector_float* r);
    virtual int row_daxpy(size_t i, float a, vector_float* r);
    virtual int    row_sum  (vector_float*);

    /// dot ( row i, col j)
    virtual float row_col_dot (size_t r, size_t c);

    virtual int dot (bool tranA, vector_float* x, vector_float* r); 
    matrix_compact_float* transpose_copy();
    /*@}*/

    /*@{*/


    // Multiplication of two matrices this = this * m or m * this
    virtual int mult (matrix_compact_float* m, int leftOrRight);

    /**
    * Subtract a sparse matrix from this matrix. The operation implemented is 
    * this := this + m.
    * @param m is a sparsematrix_compact_float* arg
    * @return 
    *   -     0 if subtraction succeeded
    *   -     1 if size mismatch
    *   -     2 if some other error
    * @warning This function can cause fill-in by destroying non-zeros, and
    * can be slow.
    */
    virtual int add (matrix_compact_float* m);


    /**
    * Subtract a sparse matrix from this matrix. The operation implemented is 
    * this := this - m.
    * @param m is a sparsematrix_compact_float* arg
    * @return 
    *   -     0 if subtraction succeeded
    *   -     1 if size mismatch
    *   -     2 if some other error
    * @warning This function can cause fill-in by destroying non-zeros, and
    * can be slow.
    */  
    virtual int sub (matrix_compact_float* m);


    /**
    * Hadamard product. After the execution of this function our current
    * matrix is: this := this .* m (in matlab notation)
    * @param m is a gsl_matrix_compact_float* argument
    * @return 
    *  -  0 if the dot product was carried out
    *  -  1 if matrix size mismatch
    *  -  2 if any other error
    */
    virtual int dot (matrix_compact_float* m);

    /**
    * Hadamard product. After the execution of this function our current
    * matrix is: this := this .* m (in matlab notation). This function does
    * a runtime type check to determine the type of 'm' to optimize the dot
    * product operation (since two ccs matrices or ccs * harwell boeing can
    * be dotted much faster without using any get(i, j) operations).
    * @param m is a sparsematrix_compact_float* argument
    * @return 
    *  -  0 if the dot product was carried out
    *  -  1 if matrix size mismatch
    *  -  2 if any other error
    */
    virtual int div (matrix_compact_float* m);


    virtual int mul (matrix_compact_float* m);
    /*@}*/

    /** @name Misc 
    * Miscellanous functions
    */
    /*@{*/

    /**
    * @return true if this matrix is external, i.e., the rows, cols
    * etc. arrays are directly initialized from outside, and not allocated
    * by us.
    */
    bool isExternal() const { return m_isExternal;}
    virtual void print ();
    virtual void transpose();

    /// Normalize the columns of the matrix to have unit L2 norm
    virtual void normalize_columns();

    virtual int compute_AtA(matrix_compact_float* result);
    virtual void error(std::string);
    /*@}*/


    /** 
     * conversions
     */
    crs_compact_float* to_crs();

  private:
    // Matrix vec mpy, result = Ax
    int matrix_vec (vector_float* x, vector_float* result);

    // Transpose[Matrix].vec = A'x
    int matrix_tran_vec (vector_float* x, vector_float* result);
  };

}

#endif // CCS_COMPACT_FLOAT
